#    Copyright 2016 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

attachment_id = '4dc3bb12-ad75-41b9-ab2c-7609e743e600'
backup_id = '707844eb-6d8a-4ac1-8b98-618e1c0b3a3a'
backup2_id = '40e8462a-c9d8-462f-a810-b732a1790535'
backup3_id = '30ae7641-017e-4221-a642-855687c8bd71'
cgsnapshot_id = '5e34cce3-bc97-46b7-a127-5cfb95ef445d'
cgsnapshot2_id = '5c36d762-d6ba-4f04-bd07-88a298cc410a'
cgsnapshot3_id = '5f392156-fc03-492a-9cb8-e46a7eedaf33'
consistency_group_id = 'f18abf73-79ee-4f2b-8d4f-1c044148f117'
image_id = 'e79161cd-5f9d-4007-8823-81a807a64332'
object_id = 'd7c5b12f-d57d-4762-99ab-db5f62ae3569'
object2_id = '51f5b8fa-c13c-48ba-8c9d-b470466cbc9c'
object3_id = '7bf5ffa9-18a2-4b64-aab4-0798b53ee4e7'
project_id = '89afd400-b646-4bbc-b12b-c0a4d63e5bd3'
project2_id = '452ebfbc-55d9-402a-87af-65061916c24b'
provider_id = '60087173-e899-470a-9e3a-ba4cffa3e3e3'
snapshot_id = '253b2878-ec60-4793-ad19-e65496ec7aab'
snapshot2_id = 'c02c44fa-5665-4a26-9e66-2ebaf25e5d2d'
snapshot3_id = '454f9970-1e05-4193-a3ed-5c390c3faa18'
user_id = 'c853ca26-e8ea-4797-8a52-ee124a013d0e'
user2_id = '95f7b7ed-bd7f-426e-b05f-f1ffeb4f09df'
volume_id = '1e5177e7-95e5-4a0f-b170-e45f4b469f6a'
volume2_id = '43a09914-e495-475f-b862-0bda3c8918e4'
volume3_id = '1b1cf149-219c-44ac-aee3-13121a7f86a7'
volume4_id = '904d4602-4301-4e9b-8df1-8133b51904e6'
volume5_id = '17b0e01d-3d2d-4c31-a1aa-c962420bc3dc'
volume_name_id = 'ee73d33c-52ed-4cb7-a8a9-2687c1205c22'
volume2_name_id = '63fbdd21-03bc-4309-b867-2893848f86af'
volume_type_id = '4e9e6d23-eed0-426d-b90a-28f87a94b6fe'
volume_type2_id = '23defc6f-6d21-4fb5-8d36-b887cbd5a19c'
will_not_be_found_id = 'ce816f65-c5aa-46d6-bd62-5272752d584a'
